/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.bu.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.tinygroup.appconfig.util.AppConfigUtil;
import org.tinygroup.bu.BusinessInstallProcessor;
import org.tinygroup.bu.BusinessUnitManager;
import org.tinygroup.bu.config.BusinessUnitInfo;
import org.tinygroup.bu.config.BusinessUnitProcessor;
import org.tinygroup.bu.config.DependencyBusinessUnit;
import org.tinygroup.bu.config.InstallProcessor;
import org.tinygroup.bu.config.UninstallProcessor;
import org.tinygroup.bu.exception.BusinessUnitRuntimeException;
import org.tinygroup.fileresolver.FileProcessor;
import org.tinygroup.logger.LogLevel;
import org.tinygroup.logger.Logger;
import org.tinygroup.logger.LoggerFactory;
import org.tinygroup.springutil.SpringUtil;
import org.tinygroup.vfs.FileObject;
import org.tinygroup.xmlparser.node.XmlNode;
import org.tinygroup.xstream.XStreamFactory;

import com.thoughtworks.xstream.XStream;

public class BusinessUnitInLibManagerImpl implements BusinessUnitManager {
	private static final String DELIMITER = "_";
	private static final String XSTEAM_PACKAGE_NAME = "bu";
	/**
	 * 公有的bu处理器
	 */
	private List<BusinessUnitProcessor> buProcessors = new ArrayList<BusinessUnitProcessor>();

	private List<BusinessUnitInfo> buInfos = new ArrayList<BusinessUnitInfo>();
	private XmlNode xmlConfig;
	private Logger logger = LoggerFactory
			.getLogger(BusinessUnitInLibManagerImpl.class);
	/**
	 * 业务单元jar与业务单元描述信息映射
	 */
	private Map<BusinessUnitInfo, FileObject> buFileMap = new HashMap<BusinessUnitInfo, FileObject>();
	/**
	 * key=package+name+version
	 */
	private Map<String, BusinessUnitInfo> nameVersionMap = new HashMap<String, BusinessUnitInfo>();
	/**
	 * 业务单元描述信息与业务单元映射
	 */
	private Map<FileObject, BusinessUnitInfo> fileBuMap = new HashMap<FileObject, BusinessUnitInfo>();
	/**
	 * key=package+name
	 */
	private Map<String, BusinessUnitInfo> nameMap = new HashMap<String, BusinessUnitInfo>();

	// 这两个现在已经没用了
	private List<FileProcessor> resourceProcessors = new ArrayList<FileProcessor>();
	private List<FileProcessor> resourceRemoveProcessors = new ArrayList<FileProcessor>();

	/**
	 * key=package+name
	 */
	private Map<String, BusinessUnitInfo> installStatus = new HashMap<String, BusinessUnitInfo>();

	public String getNodeName() {
		return "bu-config";
	}

	public void setConfiguration(XmlNode xmlNode) {
		this.xmlConfig = xmlNode;
		List<BusinessInstallProcessor> list = new ArrayList<BusinessInstallProcessor>();
		for (BusinessUnitProcessor processor : buProcessors) {
			List<InstallProcessor> installProcessors = processor
					.getInstallProcessorList();
			for (InstallProcessor installProcessor : installProcessors) {
				BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil
						.getBean(installProcessor.getProcessorBeanName());
				if (!list.contains(businessInstallProcessor))
					list.add(businessInstallProcessor);

			}
		}
		AppConfigUtil.config(this, list);
	}

	public XmlNode getConfiguration() {
		return xmlConfig;
	}

	public BusinessUnitInfo getBusinessUnitInfo(FileObject fileObject) {
		return fileBuMap.get(fileObject);
	}

	public void addBusinessUnit(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始添加业务单元文件:{0}",
				fileObject.getAbsolutePath());
		BusinessUnitInfo businessUnitInfo = instanceBusinessUnitInfo(fileObject);
		buFileMap.put(businessUnitInfo, fileObject);
		fileBuMap.put(fileObject, businessUnitInfo);
		buInfos.add(businessUnitInfo);
		nameVersionMap.put(
				getKeyVersion(businessUnitInfo.getPackageName(),
						businessUnitInfo.getName(),
						businessUnitInfo.getVersion()), businessUnitInfo);
		String key = getKey(businessUnitInfo.getPackageName(),
				businessUnitInfo.getName());
		BusinessUnitInfo info = nameMap.get(key);
		// 如果说新bu的版本比旧的高或者不存在旧的
		if (info == null
				|| businessUnitInfo.getVersion().compareTo(info.getVersion()) > 0) {
			nameMap.put(key, businessUnitInfo);
			logger.logMessage(LogLevel.INFO,
					"该业务单元文件为同类最高版本,Key:{0},Version:{1}", key,
					businessUnitInfo.getVersion());
		}
		logger.logMessage(LogLevel.INFO, "添加业务单元文件完成:{0}",
				fileObject.getAbsolutePath());
	}

	public List<BusinessUnitInfo> getBusinessUnitInfoList() {
		return buInfos;
	}

	public void install(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始安装业务单元文件:{0}",fileObject.getAbsolutePath());
		BusinessUnitInfo info = fileBuMap.get(fileObject);
		String key = getKey(info.getPackageName(), info.getName());
		logger.logMessage(LogLevel.INFO, "业务单元Key:[{0}]", key);
		if (!installStatus.containsKey(key)) {
			installBu(fileObject);
		} else {
			BusinessUnitInfo installedBu = installStatus.get(key);
			if (info.getVersion().compareTo(installedBu.getVersion()) > 1) {
				logger.logMessage(LogLevel.INFO, "业务单元Key:[{0}]已安装版本低于当前安装版本", key);
				// 此段逻辑一般不会进入，只会在以下情况下进入
				// 存在bu1的三个版本 v1 v2 v3
				// installAll时,v3被安装，但用户手动卸载v3，并手动安装v1
				// 此时若用户再手动安装v2，则v1将会被卸载，同时安装v2
				FileObject installedFile = buFileMap.get(installedBu);
				logger.logMessage(LogLevel.INFO, "开始卸载已安装版本");
				uninstall(installedFile);
				logger.logMessage(LogLevel.INFO, "卸载已安装版本完成");
				installBu(fileObject);
			}else{
				logger.logMessage(LogLevel.INFO, "业务单元Key:[{0}]已安装更高版本，请勿重新安装", key);
			}
		}
		logger.logMessage(LogLevel.INFO, "安装业务单元文件完成:{0}",fileObject.getAbsolutePath());
	}

	public void install() {
		// 由此处进入的必定是最高版本的buinfo
		for (BusinessUnitInfo info : nameMap.values()) {
			installBu(buFileMap.get(info));
		}
	}

	/**
	 * 安装bu,该方法内只进行安装，不做任何版本比较 版本比较工作需要在调用此方法之前进行处理
	 * 
	 * @param fileObject
	 */
	private void installBu(FileObject fileObject) {
		
		BusinessUnitInfo info = fileBuMap.get(fileObject);
		String key = getKey(info.getPackageName(), info.getName());
		List<FileObject> dependFileObjects = getDependBusinessUnitFileObject(fileObject);
		// 先找到其依赖的业务单元进行安装处理
		for (FileObject dependFileObject : dependFileObjects) {
			installBu(dependFileObject);
		}
		// 公共处理器进行安装处理
		publicInstallProcess(fileObject);
		installStatus.put(key, info);
	}

	public void uninstall(FileObject fileObject) {
		BusinessUnitInfo info = fileBuMap.get(fileObject);
		String key = getKey(info.getPackageName(), info.getName());
		if (!installStatus.containsKey(key)) {
			logger.logMessage(LogLevel.INFO, "业务单元Key:[{0}]未安装,无需卸载", key);
			return;
		}
		BusinessUnitInfo installedBu = installStatus.get(key);
		String installVersion = installedBu.getVersion();
		String version = info.getVersion();
		// 如果两者版本相同，则认为可以卸载了
		if ((version != null && installVersion != null && installVersion
				.equals(version))
				|| (version == null && installVersion == null)) {

		} else {
			logger.logMessage(LogLevel.INFO,
					"业务单元Key:[{0}]安装的版本{1}与当前卸载版本{2}文件不同", key, installVersion,
					version);
			return;
		}

		List<FileObject> dependByFileObjects = getDependByBusinessUnitFileObject(fileObject);
		for (FileObject dependFileObject : dependByFileObjects) {
			uninstall(dependFileObject);
		}
		// 公有处理器卸载处理
		publicUnInstallProcess(fileObject);
		installStatus.remove(fileObject);

	}

	public void uninstall(String packageName, String name, String version) {
		String versionKey = getKeyVersion(packageName, name, version);
		if (!nameVersionMap.containsKey(versionKey)) {
			return;
		}
		uninstall(buFileMap.get(nameVersionMap.get(versionKey)));
	}

	public BusinessUnitInfo getBusinessUnitInfo(String packageName,
			String name, String version) {
		return nameVersionMap.get(getKeyVersion(packageName, name, version));
	}

	public FileObject getBusinessUnit(BusinessUnitInfo businessUnitInfo) {
		return buFileMap.get(businessUnitInfo);
	}

	public void addBusinessUnitProcessor(
			BusinessUnitProcessor businessUnitProcessor) {
		buProcessors.add(businessUnitProcessor);
	}

	public List<FileProcessor> getResourceProcessors() {
		return resourceProcessors;
	}

	public List<FileProcessor> getResourceRemoveProcessors() {
		return resourceRemoveProcessors;
	}

	public void load(String path) {
		// 由于目前所有bu包均放置在lib下,所以此处无需任何实现了
	}

	private BusinessUnitInfo instanceBusinessUnitInfo(FileObject fileObject) {
		FileObject bUnit = fileObject.getChild(BUSINESS_UNIT_FILE);
		if(bUnit!=null){
			XStream stream = XStreamFactory.getXStream(XSTEAM_PACKAGE_NAME);
			logger.logMessage(LogLevel.INFO, "在[{0}]找到业务单元文件",
					fileObject.getAbsolutePath());
			return (BusinessUnitInfo) stream.fromXML(bUnit.getInputStream());
		}
		throw new BusinessUnitRuntimeException("bu.buFileNotExist",fileObject.getAbsolutePath());
	}

	private String getKeyVersion(String packageName, String name, String version) {

		return packageName + DELIMITER + name + DELIMITER + version;
	}

	private String getKey(String packageName, String name) {

		return packageName + DELIMITER + name;
	}

	/**
	 * 查询当前fileObjct(业务单元)依赖的业务单元
	 * 
	 * @param fileObject
	 * @return
	 */
	private List<FileObject> getDependBusinessUnitFileObject(
			FileObject fileObject) {
		List<FileObject> dependFileObjects = new ArrayList<FileObject>();
		BusinessUnitInfo unitInfo = getBusinessUnitInfo(fileObject);
		if (unitInfo != null) {
			List<DependencyBusinessUnit> dbUnits = unitInfo
					.getDependencyBusinessUnits();
			if (dbUnits != null) {
				for (DependencyBusinessUnit dbUnit : dbUnits) {
					BusinessUnitInfo dependUnitInfo = getBusinessUnitInfo(
							dbUnit.getPackageName(), dbUnit.getName(),
							dbUnit.getVersion());
					if (dependUnitInfo != null) {
						FileObject dependFileObject = buFileMap
								.get(dependUnitInfo);
						if (dependFileObject != null) {
							logger.logMessage(LogLevel.INFO,
									"[{0}]找到依赖的业务单元[{1}]",
									fileObject.getAbsolutePath(),
									dependFileObject.getAbsolutePath());
							dependFileObjects.add(dependFileObject);
						}
					} else {
						logger.logMessage(LogLevel.ERROR, "依赖的业务单元配置信息有错");
						throw new BusinessUnitRuntimeException(
								"[{0}]未找到依赖的业务单元文件");
					}

				}
			}
		}
		return dependFileObjects;

	}

	/**
	 * 公共处理器安装业务单元
	 * 
	 * @param fileObject
	 */
	private void publicInstallProcess(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始安装BU公有处理器");
		for (BusinessUnitProcessor processor : buProcessors) {
			processInstallBu(fileObject, processor);
		}
		logger.logMessage(LogLevel.INFO, "安装BU公有处理器完成");
	}

	/**
	 * 安装业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {
		List<InstallProcessor> installProcessors = processor
				.getInstallProcessorList();
		for (InstallProcessor installProcessor : installProcessors) {
			BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil
					.getBean(installProcessor.getProcessorBeanName());
			businessInstallProcessor.process(fileObject);
		}

	}

	private List<FileObject> getDependByBusinessUnitFileObject(
			FileObject fileObject) {
		List<FileObject> dependByFileObjects = new ArrayList<FileObject>();
		BusinessUnitInfo unitInfo = getBusinessUnitInfo(fileObject);
		if (unitInfo != null) {
			// 遍历所有安装的业务单元jar列表
			for (BusinessUnitInfo targetUnitInfo : installStatus.values()) {
				// 遍历该业务单元的依赖信息，判断是否其依赖自己，若依赖则加入列表
				for (DependencyBusinessUnit targetDepUnit : targetUnitInfo
						.getDependencyBusinessUnits()) {
					if (targetDepUnit.getName().equals(unitInfo.getName())
							&& targetDepUnit.getPackageName().equals(
									unitInfo.getPackageName())
							&& targetDepUnit.getVersion().equals(
									unitInfo.getVersion())) {
						dependByFileObjects.add(buFileMap.get(targetUnitInfo));
					}
				}
			}
		}
		return dependByFileObjects;
	}

	private void publicUnInstallProcess(FileObject fileObject) {
		if (buProcessors != null) {
			for (BusinessUnitProcessor processor : buProcessors) {

				processUnInstallBu(fileObject, processor);
			}
		}
	}

	/**
	 * 卸载业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processUnInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {

		List<UninstallProcessor> uninstallProcessors = processor
				.getUninstallProcessorList();
		if (uninstallProcessors != null) {
			for (UninstallProcessor uninstallProcessor : uninstallProcessors) {
				BusinessInstallProcessor businessUnInstallProcessor = (BusinessInstallProcessor) SpringUtil
						.getBean(uninstallProcessor.getProcessorBeanName());
				businessUnInstallProcessor.process(fileObject);
			}
		}
	}
}
