/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.bu.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.tinygroup.appconfig.util.AppConfigUtil;
import org.tinygroup.bu.BusinessInstallProcessor;
import org.tinygroup.bu.BusinessUnitManager;
import org.tinygroup.bu.config.BusinessUnitInfo;
import org.tinygroup.bu.config.BusinessUnitProcessor;
import org.tinygroup.bu.config.DependencyBusinessUnit;
import org.tinygroup.bu.config.InstallProcessor;
import org.tinygroup.bu.config.UninstallProcessor;
import org.tinygroup.bu.exception.BusinessUnitRuntimeException;
import org.tinygroup.bu.loader.BusinessUnitLoader;
import org.tinygroup.bu.loader.impl.BusinessUnitLibLoaderImpl;
import org.tinygroup.commons.tools.Assert;
import org.tinygroup.fileresolver.FileProcessor;
import org.tinygroup.logger.LogLevel;
import org.tinygroup.logger.Logger;
import org.tinygroup.logger.LoggerFactory;
import org.tinygroup.springutil.SpringUtil;
import org.tinygroup.vfs.FileObject;
import org.tinygroup.xmlparser.node.XmlNode;
import org.tinygroup.xstream.XStreamFactory;

import com.thoughtworks.xstream.XStream;

/**
 * 业务单元管理器
 * 
 * @author renhui
 * 
 */
public class BusinessUnitManagerImpl implements BusinessUnitManager {
	private static Logger logger = LoggerFactory
			.getLogger(BusinessUnitManagerImpl.class);
	private static final String DELIMITER = "_";
	private static final String XSTEAM_PACKAGE_NAME = "bu";
	private XmlNode xmlConfig;
	private List<BusinessUnitInfo> businessUnitInfos = new ArrayList<BusinessUnitInfo>();
	/**
	 * key=package+name+version
	 */
	private Map<String, BusinessUnitInfo> nameVersionMap = new HashMap<String, BusinessUnitInfo>();

	/**
	 * key=package+name
	 */
	private Map<String, List<BusinessUnitInfo>> nameMap = new HashMap<String, List<BusinessUnitInfo>>();

	/**
	 * 公有的bu处理器
	 */
	private List<BusinessUnitProcessor> publicBup = new ArrayList<BusinessUnitProcessor>();
	/**
	 * 私有的bu处理器
	 */
	private Map<FileObject, BusinessUnitProcessor> privateBup = new HashMap<FileObject, BusinessUnitProcessor>();

	/**
	 * 业务单元jar与业务单元描述信息映射
	 */
	private Map<BusinessUnitInfo, FileObject> buFileMap = new HashMap<BusinessUnitInfo, FileObject>();

	/**
	 * 业务单元描述信息与业务单元映射
	 */
	private Map<FileObject, BusinessUnitInfo> fileBuMap = new HashMap<FileObject, BusinessUnitInfo>();

	/**
	 * 业务单元文件是否被安装
	 */
	private Map<FileObject, Boolean> installStatus = new HashMap<FileObject, Boolean>();

	private List<FileProcessor> resourceProcessors = new ArrayList<FileProcessor>();
	private List<FileProcessor> resourceRemoveProcessors = new ArrayList<FileProcessor>();
	private BusinessUnitLoader loader = new BusinessUnitLibLoaderImpl();

	public BusinessUnitInfo getBusinessUnitInfo(FileObject fileObject) {
		return fileBuMap.get(fileObject);
	}

	public FileObject getBusinessUnit(BusinessUnitInfo businessUnitInfo) {
		return buFileMap.get(businessUnitInfo);
	}

	public void addBusinessUnit(FileObject fileObject) {
		Assert.assertNotNull(fileObject, "fileObject must not null");
		BusinessUnitInfo businessUnitInfo = instanceBusinessUnitInfo(fileObject);
		buFileMap.put(businessUnitInfo, fileObject);
		fileBuMap.put(fileObject, businessUnitInfo);
		businessUnitInfos.add(businessUnitInfo);
		nameVersionMap.put(
				getKeyVersion(businessUnitInfo.getPackageName(),
						businessUnitInfo.getName(),
						businessUnitInfo.getVersion()), businessUnitInfo);
		String key = getKey(businessUnitInfo.getPackageName(),
				businessUnitInfo.getName());
		List<BusinessUnitInfo> infos = nameMap.get(key);
		if (infos == null) {
			infos = new ArrayList<BusinessUnitInfo>();
		}
		infos.add(businessUnitInfo);
		nameMap.put(key, infos);
	}

	private BusinessUnitInfo instanceBusinessUnitInfo(FileObject fileObject) {
		FileObject bUnit = fileObject.getChild(BUSINESS_UNIT_FILE);
		if(bUnit!=null){
			XStream stream = XStreamFactory.getXStream(XSTEAM_PACKAGE_NAME);
			logger.logMessage(LogLevel.INFO, "在[{0}]找到业务单元文件",
					fileObject.getAbsolutePath());
			return (BusinessUnitInfo) stream.fromXML(bUnit.getInputStream());
		}
		throw new BusinessUnitRuntimeException("bu.buFileNotExist",fileObject.getAbsolutePath());
	}

	public List<BusinessUnitInfo> getBusinessUnitInfoList() {
		return businessUnitInfos;
	}

	public void install(FileObject fileObject) {
		loadResource(fileObject);
		installBu(fileObject);
	}

	private void loadResource(FileObject fileObject) {
		// 扫描资源
		try {
			loader.loadBuInfo(fileObject, fileBuMap.get(fileObject));
		} catch (Exception e) {
			logger.errorMessage("安装时，加载业务单元资源时出错，文件:{0}", e, fileObject);
		}
	}

	private void installBu(FileObject fileObject) {

		logger.logMessage(LogLevel.INFO, "开始安装业务单元[{0}]",
				fileObject.getAbsolutePath());
		if (!installStatus.containsKey(fileObject)
				|| !installStatus.get(fileObject)) {
			List<FileObject> dependFileObjects = getDependBusinessUnitFileObject(fileObject);
			// 先找到其依赖的业务单元进行安装处理
			for (FileObject dependFileObject : dependFileObjects) {
				installBu(dependFileObject);
			}

			// 先卸载原有版本的
			unstallOriginalVersion(fileObject);
			// 公共处理器进行安装处理
			publicInstallProcess(fileObject);
			// 私有处理器进行安装处理
			privateInstallProcess(fileObject);

			installStatus.put(fileObject, true);
		} else {
			logger.logMessage(LogLevel.INFO, "业务单元[{0}]已安装,请勿重新安装",
					fileObject.getAbsolutePath());
		}

		logger.logMessage(LogLevel.INFO, "安装业务单元[{0}]结束",
				fileObject.getAbsolutePath());

	}

	/**
	 * 查询当前fileObjct(业务单元)依赖的业务单元
	 * 
	 * @param fileObject
	 * @return
	 */
	private List<FileObject> getDependBusinessUnitFileObject(
			FileObject fileObject) {
		List<FileObject> dependFileObjects = new ArrayList<FileObject>();
		BusinessUnitInfo unitInfo = getBusinessUnitInfo(fileObject);
		if (unitInfo != null) {
			List<DependencyBusinessUnit> dbUnits = unitInfo
					.getDependencyBusinessUnits();
			if (dbUnits != null) {
				for (DependencyBusinessUnit dbUnit : dbUnits) {
					BusinessUnitInfo dependUnitInfo = getBusinessUnitInfo(
							dbUnit.getPackageName(), dbUnit.getName(),
							dbUnit.getVersion());
					if (dependUnitInfo != null) {
						FileObject dependFileObject = buFileMap
								.get(dependUnitInfo);
						if (dependFileObject != null) {
							logger.logMessage(LogLevel.INFO,
									"[{0}]找到依赖的业务单元[{1}]",
									fileObject.getAbsolutePath(),
									dependFileObject.getAbsolutePath());
							dependFileObjects.add(dependFileObject);
						}
					} else {
						logger.logMessage(LogLevel.ERROR, "依赖的业务单元配置信息有错");
						throw new BusinessUnitRuntimeException(
								"[{0}]未找到依赖的业务单元文件");
					}

				}
			}
		}
		return dependFileObjects;

	}

	/**
	 * 查询依赖当前fileObject(业务单元)的业务单元
	 * 
	 * @param fileObject
	 * @return
	 */
	private List<FileObject> getDependByBusinessUnitFileObject(
			FileObject fileObject) {
		List<FileObject> dependByFileObjects = new ArrayList<FileObject>();
		BusinessUnitInfo unitInfo = getBusinessUnitInfo(fileObject);
		if (unitInfo != null) {
			// 遍历所有安装的业务单元jar列表
			for (FileObject targetFile : installStatus.keySet()) {
				if (!installStatus.get(targetFile)) {
					continue;
				}
				// 获取已安装的业务单元信息
				BusinessUnitInfo targetUnitInfo = getBusinessUnitInfo(targetFile);
				// 遍历该业务单元的依赖信息，判断是否其依赖自己，若依赖则加入列表
				for (DependencyBusinessUnit targetDepUnit : targetUnitInfo
						.getDependencyBusinessUnits()) {
					if (targetDepUnit.getName().equals(unitInfo.getName())
							&& targetDepUnit.getPackageName().equals(
									unitInfo.getPackageName())
							&& targetDepUnit.getVersion().equals(
									unitInfo.getVersion())) {
						dependByFileObjects.add(targetFile);
					}
				}
			}
		}
		return dependByFileObjects;
	}

	public void uninstall(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始卸载业务单元[{0}]",
				fileObject.getAbsolutePath());

		if (installStatus.containsKey(fileObject)
				&& installStatus.get(fileObject)) {
			List<FileObject> dependByFileObjects = getDependByBusinessUnitFileObject(fileObject);
			for (FileObject dependFileObject : dependByFileObjects) {
				uninstall(dependFileObject);
			}
			// 20130114chenjiao添加
			BusinessUnitInfo info = fileBuMap.remove(fileObject);
			buFileMap.remove(info);

			// 私有处理器卸载处理
			privateUnInstallProcess(fileObject);
			// 公有处理器卸载处理
			publicUnInstallProcess(fileObject);
			// 删除扫描所得资源
			loader.remove(fileObject);
			installStatus.remove(fileObject);
		} else {
			logger.logMessage(LogLevel.INFO, "业务单元[{0}]未被安装,无需卸载",
					fileObject.getAbsolutePath());
		}

		logger.logMessage(LogLevel.INFO, "卸载业务单元[{0}]结束",
				fileObject.getAbsolutePath());
	}

	private void privateInstallProcess(FileObject fileObject) {
		// 需要在安装过程中查找jar包中私有的处理器
		BusinessUnitProcessor eachProcessor = getEachProcessor(fileObject);
		if (eachProcessor != null)
			processInstallBu(fileObject, eachProcessor);

	}

	private void privateUnInstallProcess(FileObject fileObject) {
		// 需要在安装过程中查找jar包中私有的处理器
		BusinessUnitProcessor eachProcessor = getEachProcessor(fileObject);
		if (eachProcessor != null) {
			processUnInstallBu(fileObject, eachProcessor);
			// 20130114chenjiao添加
			privateBup.remove(fileObject);
		}

	}

	/**
	 * 公共处理器安装业务单元
	 * 
	 * @param fileObject
	 */
	private void publicInstallProcess(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始安装BU公有处理器");
		for (BusinessUnitProcessor processor : publicBup) {
			processInstallBu(fileObject, processor);
		}
		logger.logMessage(LogLevel.INFO, "安装BU公有处理器完成");
	}

	/**
	 * 公共处理器安装业务单元
	 * 
	 * @param fileObject
	 */
	private void publicUnInstallProcess(FileObject fileObject) {
		if (publicBup != null) {
			for (BusinessUnitProcessor processor : publicBup) {

				processUnInstallBu(fileObject, processor);
			}
		}
	}

	/**
	 * 安装业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {

		List<InstallProcessor> installProcessors = processor
				.getInstallProcessorList();
		for (InstallProcessor installProcessor : installProcessors) {
			BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil.getBean(
							installProcessor.getProcessorBeanName());
			businessInstallProcessor.process(fileObject);
		}

	}

	/**
	 * 卸载业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processUnInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {

		List<UninstallProcessor> uninstallProcessors = processor
				.getUninstallProcessorList();
		if (uninstallProcessors != null) {
			for (UninstallProcessor uninstallProcessor : uninstallProcessors) {
				BusinessInstallProcessor businessUnInstallProcessor = (BusinessInstallProcessor) SpringUtil.getBean(
								uninstallProcessor.getProcessorBeanName());
				businessUnInstallProcessor.process(fileObject);
			}
		}
	}

	/**
	 * 卸载原有版本
	 * 
	 * @param fileObject
	 */
	private void unstallOriginalVersion(FileObject fileObject) {

		BusinessUnitInfo businessUnitInfo = fileBuMap.get(fileObject);
		if (businessUnitInfo != null) {
			String key = getKey(businessUnitInfo.getPackageName(),
					businessUnitInfo.getName());
			List<BusinessUnitInfo> originalUnitInfos = nameMap.get(key);
			if (originalUnitInfos != null) {
				for (BusinessUnitInfo originalUnitInfo : originalUnitInfos) {
					if (businessUnitInfo != null
							&& businessUnitInfo != originalUnitInfo) {
						FileObject original = buFileMap.get(originalUnitInfo);
						if (original != null) {
							uninstall(original);
						}
					}
				}
			}

		}

	}

	private BusinessUnitProcessor getEachProcessor(FileObject fileObject) {
		BusinessUnitProcessor businessUnitProcessor = privateBup
				.get(fileObject);
		if (businessUnitProcessor == null) {
			businessUnitProcessor = newBusinessUnitProcessor(fileObject);
			if (businessUnitProcessor != null)
				privateBup.put(fileObject, businessUnitProcessor);
		}
		return businessUnitProcessor;
	}

	private BusinessUnitProcessor newBusinessUnitProcessor(FileObject fileObject) {
		BusinessUnitProcessor businessUnitProcessor = null;
		FileObject processorFile = fileObject
				.getChild(BUSINESS_UNIT_PROCESS_FILE);
		if (processorFile == null) {
			logger.logMessage(LogLevel.INFO, "[{0}]无自定义的业务单元处理器",
					fileObject.getAbsolutePath());
			return businessUnitProcessor;
		}
		XStream stream = XStreamFactory.getXStream(XSTEAM_PACKAGE_NAME);
		logger.logMessage(LogLevel.INFO, "在[{0}]找到自定义的业务单元处理器",
				fileObject.getAbsolutePath());
		businessUnitProcessor = (BusinessUnitProcessor) stream
				.fromXML(processorFile.getInputStream());
		return businessUnitProcessor;
	}

	public BusinessUnitInfo getBusinessUnitInfo(String packageName,
			String name, String version) {
		return nameVersionMap.get(getKeyVersion(packageName, name, version));
	}

	/**
	 * 增加业务处理器
	 */
	public void addBusinessUnitProcessor(
			BusinessUnitProcessor businessUnitProcessor) {
		publicBup.add(businessUnitProcessor);
	}

	private String getKeyVersion(String packageName, String name, String version) {

		return packageName + DELIMITER + name + DELIMITER + version;
	}

	private String getKey(String packageName, String name) {

		return packageName + DELIMITER + name;
	}

	public List<FileProcessor> getResourceProcessors() {
		return resourceProcessors;
	}

	public List<FileProcessor> getResourceRemoveProcessors() {
		return resourceRemoveProcessors;
	}

	public void load(String path) {
		loader.load(path);
	}

	public void uninstall(String packageName, String name, String version) {
		BusinessUnitInfo info = getBusinessUnitInfo(packageName, name, version);
		FileObject file = buFileMap.get(info);
		uninstall(file);
	}

	public void install() {
		for (FileObject file : buFileMap.values()) {
			install(file);
		}
	}

	public String getNodeName() {
		return "bu-config";
	}

	public void setConfiguration(XmlNode xmlNode) {
		this.xmlConfig = xmlNode;
		List<BusinessInstallProcessor> list = new ArrayList<BusinessInstallProcessor>();
		for (BusinessUnitProcessor processor : publicBup) {
			List<InstallProcessor> installProcessors = processor
					.getInstallProcessorList();
			for (InstallProcessor installProcessor : installProcessors) {
				BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil.getBean(
								installProcessor.getProcessorBeanName());
				if (!list.contains(businessInstallProcessor))
					list.add(businessInstallProcessor);

			}
		}
		for (BusinessUnitProcessor processor : privateBup.values()) {
			List<InstallProcessor> installProcessors = processor
					.getInstallProcessorList();
			for (InstallProcessor installProcessor : installProcessors) {
				BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil.getBean(
								installProcessor.getProcessorBeanName());
				if (!list.contains(businessInstallProcessor))
					list.add(businessInstallProcessor);
			}
		}
		AppConfigUtil.config(this, list);
	}

	public XmlNode getConfiguration() {
		return xmlConfig;
	}

}
