/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.bu.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.tinygroup.appconfig.util.AppConfigUtil;
import org.tinygroup.bu.BusinessInstallProcessor;
import org.tinygroup.bu.BusinessUnitManager;
import org.tinygroup.bu.config.BusinessUnitInfo;
import org.tinygroup.bu.config.BusinessUnitProcessor;
import org.tinygroup.bu.config.InstallProcessor;
import org.tinygroup.bu.config.UninstallProcessor;
import org.tinygroup.bu.exception.BusinessUnitRuntimeException;
import org.tinygroup.commons.tools.Assert;
import org.tinygroup.fileresolver.FileProcessor;
import org.tinygroup.logger.LogLevel;
import org.tinygroup.logger.Logger;
import org.tinygroup.logger.LoggerFactory;
import org.tinygroup.springutil.SpringUtil;
import org.tinygroup.vfs.FileObject;
import org.tinygroup.xmlparser.node.XmlNode;
import org.tinygroup.xstream.XStreamFactory;

import com.thoughtworks.xstream.XStream;

public class SimpleBusinessUnitInLibManagerImpl implements BusinessUnitManager {
	private static final String DELIMITER = "_";
	private static final String XSTEAM_PACKAGE_NAME = "bu";
	/**
	 * 公有的bu处理器
	 */
	private List<BusinessUnitProcessor> buProcessors = new ArrayList<BusinessUnitProcessor>();

	private List<BusinessUnitInfo> buInfos = new ArrayList<BusinessUnitInfo>();
	private XmlNode xmlConfig;
	private Logger logger = LoggerFactory
			.getLogger(SimpleBusinessUnitInLibManagerImpl.class);
	/**
	 * 业务单元jar与业务单元描述信息映射
	 */
	private Map<BusinessUnitInfo, FileObject> buFileMap = new HashMap<BusinessUnitInfo, FileObject>();
	/**
	 * key=package+name+version
	 */
	private Map<String, BusinessUnitInfo> nameVersionMap = new HashMap<String, BusinessUnitInfo>();
	/**
	 * 业务单元描述信息与业务单元映射
	 */
	private Map<FileObject, BusinessUnitInfo> fileBuMap = new HashMap<FileObject, BusinessUnitInfo>();
	/**
	 * key=package+name
	 */
	// 这两个现在已经没用了
	private List<FileProcessor> resourceProcessors = new ArrayList<FileProcessor>();
	private List<FileProcessor> resourceRemoveProcessors = new ArrayList<FileProcessor>();

	public String getNodeName() {
		return "bu-config";
	}

	public void setConfiguration(XmlNode xmlNode) {
		this.xmlConfig = xmlNode;
		List<BusinessInstallProcessor> list = new ArrayList<BusinessInstallProcessor>();
		for (BusinessUnitProcessor processor : buProcessors) {
			for (InstallProcessor installProcessor : processor.getInstallProcessorList()) {
				BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil
						.getBean(installProcessor.getProcessorBeanName());
				if (!list.contains(businessInstallProcessor))
					list.add(businessInstallProcessor);

			}
			for (UninstallProcessor uninstallProcessor : processor.getUninstallProcessorList()) {
				BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil
						.getBean(uninstallProcessor.getProcessorBeanName());
				if (!list.contains(businessInstallProcessor))
					list.add(businessInstallProcessor);

			}

		}
		AppConfigUtil.config(this, list);
	}

	public XmlNode getConfiguration() {
		return xmlConfig;
	}

	public BusinessUnitInfo getBusinessUnitInfo(FileObject fileObject) {
		return fileBuMap.get(fileObject);
	}

	public void addBusinessUnit(FileObject fileObject) {
		Assert.assertNotNull(fileObject, "fileObject must not null");
		logger.logMessage(LogLevel.INFO, "开始添加业务单元文件:{0}",
				fileObject.getAbsolutePath());
		BusinessUnitInfo businessUnitInfo = instanceBusinessUnitInfo(fileObject);
		buFileMap.put(businessUnitInfo, fileObject);
		fileBuMap.put(fileObject, businessUnitInfo);
		buInfos.add(businessUnitInfo);
		nameVersionMap.put(
				getKeyVersion(businessUnitInfo.getPackageName(),
						businessUnitInfo.getName(),
						businessUnitInfo.getVersion()), businessUnitInfo);
		logger.logMessage(LogLevel.INFO, "添加业务单元文件完成:{0}",
				fileObject.getAbsolutePath());
	}

	public List<BusinessUnitInfo> getBusinessUnitInfoList() {
		return buInfos;
	}

	public void install(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始安装BU文件:{0}",fileObject.getAbsolutePath());
		// 公共处理器进行安装处理
		publicInstallProcess(fileObject);
		logger.logMessage(LogLevel.INFO, "安装BU文件完成:{0}",fileObject.getAbsolutePath());
	}

	public void install() {
		for (FileObject file : fileBuMap.keySet()) {
			install(file);
		}
	}

	public void uninstall(FileObject fileObject) {
		// 公有处理器卸载处理
		publicUnInstallProcess(fileObject);
	}

	public void uninstall(String packageName, String name, String version) {
		String versionKey = getKeyVersion(packageName, name, version);
		if (!nameVersionMap.containsKey(versionKey)) {
			return;
		}
		uninstall(buFileMap.get(nameVersionMap.get(versionKey)));
	}

	public BusinessUnitInfo getBusinessUnitInfo(String packageName,
			String name, String version) {
		return nameVersionMap.get(getKeyVersion(packageName, name, version));
	}

	public FileObject getBusinessUnit(BusinessUnitInfo businessUnitInfo) {
		return buFileMap.get(businessUnitInfo);
	}

	public void addBusinessUnitProcessor(
			BusinessUnitProcessor businessUnitProcessor) {
		buProcessors.add(businessUnitProcessor);
	}

	public List<FileProcessor> getResourceProcessors() {
		return resourceProcessors;
	}

	public List<FileProcessor> getResourceRemoveProcessors() {
		return resourceRemoveProcessors;
	}

	public void load(String path) {
		// 由于目前所有bu包均放置在lib下,所以此处无需任何实现了
	}

	private BusinessUnitInfo instanceBusinessUnitInfo(FileObject fileObject) {
		FileObject bUnit = fileObject.getChild(BUSINESS_UNIT_FILE);
		if(bUnit!=null){
			XStream stream = XStreamFactory.getXStream(XSTEAM_PACKAGE_NAME);
			logger.logMessage(LogLevel.INFO, "在[{0}]找到业务单元文件",
					fileObject.getAbsolutePath());
			return (BusinessUnitInfo) stream.fromXML(bUnit.getInputStream());
		}
		throw new BusinessUnitRuntimeException("bu.buFileNotExist",fileObject.getAbsolutePath());
	}

	private String getKeyVersion(String packageName, String name, String version) {

		return packageName + DELIMITER + name + DELIMITER + version;
	}

	/**
	 * 公共处理器安装业务单元
	 * 
	 * @param fileObject
	 */
	private void publicInstallProcess(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始安装BU公有处理器");
		for (BusinessUnitProcessor processor : buProcessors) {
			processInstallBu(fileObject, processor);
		}
		logger.logMessage(LogLevel.INFO, "安装BU公有处理器完成");
	}

	/**
	 * 安装业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {
		List<InstallProcessor> installProcessors = processor
				.getInstallProcessorList();
		for (InstallProcessor installProcessor : installProcessors) {
			BusinessInstallProcessor businessInstallProcessor = (BusinessInstallProcessor) SpringUtil
					.getBean(installProcessor.getProcessorBeanName());
			businessInstallProcessor.process(fileObject);
		}

	}

	private void publicUnInstallProcess(FileObject fileObject) {
		logger.logMessage(LogLevel.INFO, "开始卸载BU公有处理器");
		for (BusinessUnitProcessor processor : buProcessors) {
			processUnInstallBu(fileObject, processor);
		}
		logger.logMessage(LogLevel.INFO, "卸载BU公有处理器完成");
	}

	/**
	 * 卸载业务单元处理
	 * 
	 * @param fileObject
	 * @param processor
	 */
	private void processUnInstallBu(FileObject fileObject,
			BusinessUnitProcessor processor) {

		List<UninstallProcessor> uninstallProcessors = processor
				.getUninstallProcessorList();
		if (uninstallProcessors != null) {
			for (UninstallProcessor uninstallProcessor : uninstallProcessors) {
				BusinessInstallProcessor businessUnInstallProcessor = (BusinessInstallProcessor) SpringUtil
						.getBean(uninstallProcessor.getProcessorBeanName());
				businessUnInstallProcessor.process(fileObject);
			}
		}
	}
}
